' basic module for language training Spanish-German
#include once "fbthread.bi"
dim shared as any ptr speakLock

const MODIFY_TOLOWER  =  1, MODIFY_TOUPPER =  2, MODIFY_NOPARANTHESES =   4, MODIFY_NOSPECIAL = 8, MODIFY_NOACUTE = 16
const MODIFY_NOUMLAUT = 32, MODIFY_NOTILDE = 64, MODIFY_SORTABLE      = 128

namespace INFLECT
  const PERS1 = 1, PERS2 = 2, PERS3 = 3
  const SINGULAR = 0, PLURAL = 3
  const PRES = 0, PRETIND = 8, PRETIMP = 16, PERF = 24, PQPERF = 32
  const FUTSIMP = 40, FUTCOMP = 48, IMPERAT = 56
end namespace

' misc
const COLOR_GRAY = rgb(127, 127, 127)

enum GUI
  none = 0, enter, speech, repeat, level, keys = level+B_COUNT
end enum

declare sub arrayToList(l as TList, lpArray() as TListPoint)
declare function conjugation(verb as TWord, inflection as integer) as TWord
declare function number(first as integer = 0, last as integer = 0) as TWord
declare function chooseWords(exercise as TExercise, opt as long = 1) as long
declare sub getKeyArray(key(any) as string)
declare sub guiAddKeys(byref tempGui as TGui, sID as integer, keys() as string, sx as integer, sy as integer, _
                       w as integer, h as integer, dx as integer, dy as integer = 0, perLine as integer = 255)
declare sub guiAddLevs(byref tempGui as TGui, sID as integer, sx as integer, sy as integer, _
                       w as integer, h as integer, dx as integer, fontsize as integer = 14)
declare sub init
declare function infoDefault(exercise as TExercise) as string
declare function infoEmpty(exercise as TExercise) as string
declare sub listToArray(l as TList, key as string, keyArray() as string, lpArray() as TListPoint)
declare sub mainWindow
declare function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
declare function searchInput(txt as string = "", x as integer = SCREEN_X\2 - 250, y as integer = SCREEN_Y\2-150, _
                             w as integer = 500, h as integer = 300) as string
declare function simpleTest(exercice as TExercise) as long
declare sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
declare sub speakThread(txt as string, no as integer = 1)
declare sub wordInfo(w as TWord)

init
dim as TList fullList
fullList.listname = "full"
fullList.add , 0, &hffffff
fullList.del 0, &hffffff, "group", "cardinal"
fullList.save

#include once "learning.bi"
globalExercises.add(TExercise("Spanisch -> Deutsch", "default-es-de",  20, _
                    @simpleTest, @infoDefault, @chooseWords))
globalExercises.add(TExercise("Deutsch -> Spanisch", "default-de-es", -20, _
                    @simpleTest, @infoDefault, @chooseWords))
#include once "crossword.bi"
#include once "dictionary.bi"
#include once "export.bi"

' update lists
scope
  dim as TList list(1)
  dim as TListPoint listPoint
  dim as integer size
  for i as integer = 0 to 1
    with list(i)
      .listName = "default-" & iif(i = 0, "es-de", "de-es")
      .list.addAll(fullList.list)
      .load
      size = fullList.list.size
      .sort
      .mix
      .save
      dim as integer k = 0
      foreach(TListPoint, listPoint, .list)
        dim as TWord tempWord = globallist.get(listPoint.id)
        if tempWord.id then
          if i = 0 then
            tempWord.levelA = listPoint.factor*10 - 12
          else
            tempWord.levelB = listPoint.factor*10 - 12
          end if
          globallist.put(listPoint.id, tempWord)
        else
          fullList.list.remove(listPoint)
        end if
        k += 1
        drawProgress(100, SCREEN_Y\2-10, SCREEN_X-200, 20, size*i+k, size*2)
      nexteach
      .list.clear
    end with
  next
end scope
#include once "pairs.bi"
#include once "similar.bi"
#include once "conjugation.bi"

sub arrayToList(l as TList, lpArray() as TListPoint)
  l.list.clear
  for i as integer = 0 to ubound(lpArray)
    l.list.add(lpArray(i))
  next
end sub

function chooseWords(exercise as TExercise, opt as long = 1) as long
  dim as integer lXstart =  20, lYstart =  80, lWidth =  60, lHeight = 22
  dim as integer lXplus  =  10, lYplus  =   5, lMod   =  11, lLines  = 10
  dim as integer tXstart =  20, tYstart = 380, tWidth = 140, tHeight = 22
  dim as integer tXplus  =  10, tYplus  =   5, tMod   =   5, tLines  =  8
  dim as integer mXstart = 140, mYStart =  20, mWidth =  40, mHeight = 22
  dim as integer mXPlus  =  20, mYPlus  =   0, mMod   =   8, mLines  =  1
  dim as integer mIndex = 50, fontsize = 12
  dim as integer okX = 650, okY = 600, okW = 120, okH = 25, okIndex = 99
  dim as integer chooseAllX = 30, chooseAllY = okY, chooseAllW = okW, chooseAllH = okH, chooseAllIndex = 98
  dim as integer chooseNoX = 200, chooseNoY  = okY, chooseNoW  = okW, chooseNoH  = okH, chooseNoIndex  = 97
  dim as TGuiButton tempButton

  dim as integer mx, my, mb, i
  dim as long lession, lCount = 0, tCount = 0
  dim as string key, tagname, tags(tMod*tLines-1)
  dim as mdList(long) ks1 = listOfLessions.keySet
  dim as TGui settingGui
  dim as TGuiEvent event

  if len(exercise.settingString) < lMod*lLines + tMod*tLines + 1 then
    exercise.settingString &= string((lMod*lLines + tMod*tLines + 1) - len(exercise.settingString), 0)
  end if
  if opt <> 0 then
    cls
    printFT 20, 50, "Wörter aus folgenden Kapiteln:", font, 16
    printFT 20, 340, "Wörter aus folgenden Wortgruppen:", font, 16
    printFT 20, mYStart, "Markierung:", font, 16
  end if

  ' create GUI
  i = 0
  foreach(long, lession, ks1)
    if lession >= 100 then
      dim as TGuiButton tempButton = TGuiButton(lession, lXstart+(lCount mod lMod)*(lWidth+lXplus), _
             lYStart+(lCount\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
      if exercise.settingString[i+1] > 0 then
        tempButton.state = GuiState.activated
      end if
      tempButton.fontsize = fontsize
      settingGui.buttons.add tempButton
      lCount += 1
      if lCount = lMod*lLines then exit while
    end if
    i += 1
  nexteach
  i = 0
  dim as mdList(string) ks2 = listOfTags.keySet
  foreach(string, tagname, ks2)
    if listOfTags.get(tagname) > 3 then
      tags(tCount) = tagname
      dim as TGuiButton tempButton = TGuiButton(tCount+1, tXstart+(tCount mod tMod)*(tWidth+tXplus), _
             tYStart+(tCount\tMod)*(tHeight+tYplus), tWidth, tHeight, tagname)
      if exercise.settingString[lMod*lLines+tCount+1] > 0 then
        tempButton.state = GuiState.activated
      end if
      tempButton.fontsize = fontsize
      settingGui.buttons.add tempButton
      tCount += 1
      if tCount = tMod*tLines then exit while
    end if
    i += 1
  nexteach
  i = 0
  for i as integer = 0 to 7
    tempButton = TGuiButton(i+mIndex, mXstart+i*(mWidth+mXplus), mYStart, mWidth, mHeight, str(i+1))
    if bit(exercise.settingString[0], i) then tempButton.state = GuiState.activated
    tempButton.fontsize = fontsize
    settingGui.buttons.add tempButton
  next
  tempButton = TGuiButton(okIndex, okX, okY, okW, okH, "zurück", chr(27))
  tempButton.fontsize = fontsize
  settingGui.buttons.add tempButton
  tempButton = TGuiButton(chooseNoIndex,  chooseNoX,  chooseNoY,  chooseNoW,  chooseNoH, "alle abwählen")
  tempButton.fontsize = fontsize
  settingGui.buttons.add(tempButton)
  tempButton = TGuiButton(chooseAllIndex, chooseAllX, chooseAllY, chooseAllW, chooseAllH, "alle Kapitel")
  tempButton.fontsize = fontsize
  settingGui.buttons.add(tempButton)
  ' check input
  if opt <> 0 then
    settingGui.show
    do
      getmouse mx, my,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      event = settingGui.getEvent(key)
      if event.mbutton = 1 orelse len(event.key) then
        select case event.id
          case okIndex
          case chooseNoIndex
            ' toggle off marks
            for i as integer = 0 to 7
              tempButton = TGuiButton(i+mIndex, mXstart+i*(mWidth+mXplus), mYStart, mWidth, mHeight, str(i+1))
              tempButton.state = GuiState.active
              tempButton.fontsize = fontsize
              settingGui.setButton tempButton
            next
            ' toggle off lessions
            dim as integer i = 0
            foreach(long, lession, ks1)
              if lession >= 100 then
                dim as TGuiButton tempButton = TGuiButton(lession, lXstart+(i mod lMod)*(lWidth+lXplus), _
                       lYStart+(i\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
                tempButton.state = GuiState.active
                tempButton.fontsize = fontsize
                settingGui.setButton tempButton
                i += 1
                if i >= lCount then exit while
              end if
            nexteach
            i = 0
            ' toggle off tags
            foreach(string, tagname, ks2)
              if listOfTags.get(tagname) > 3 then
                tags(tCount) = tagname
                dim as TGuiButton tempButton = TGuiButton(i+1, tXstart+(i mod tMod)*(tWidth+tXplus), _
                       tYStart+(i\tMod)*(tHeight+tYplus), tWidth, tHeight, tagname)
                tempButton.state = GuiState.active
                tempButton.fontsize = fontsize
                settingGui.setButton tempButton
                i += 1
                if i >= lCount then exit while
              end if
            nexteach
            settingGui.show
          case chooseAllIndex
            ' toggle on lessions
            dim as integer i = 0
            foreach(long, lession, ks1)
              if lession >= 100 then
                dim as TGuiButton tempButton = TGuiButton(lession, lXstart+(i mod lMod)*(lWidth+lXplus), _
                       lYStart+(i\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
                tempButton.state = GuiState.activated
                tempButton.fontsize = fontsize
                settingGui.setButton tempButton
                i += 1
                if i >= lCount then exit while
              end if
            nexteach
            settingGui.show
          case else
            dim as TGuiButton tempButton = settingGui.getButton(event.id)
            tempButton.state xor= GuiState.activated
            settingGui.setButton(tempButton)
            tempButton.show
        end select
      end if
      do
        getmouse mx, my,, mb
        sleep 1
      loop until mb < 1
      if event.id = okIndex and (key = chr(27) or checkMouseIn(okX, okY, okW, okH) <> 0) then exit do
    loop until key = chr(255, 107)
  end if

  ' create exercise.currentList
  exercise.currentList.list.clear
  exercise.settingString = string((lMod*lLines + tMod*tLines + 1), 0)
  dim lessions as mdList(long), countKey as integer = 0
  redim as string keys(ubound(tags)+8), values(ubound(tags)+8)
  foreach(TGuiButton, tempButton, settingGui.buttons)
    if tempButton.state and GuiState.activated then
      select case tempButton.id
        case 1 to tCount
          keys(countKey) = "tag"
          values(countKey) = tags(tempButton.id-1)
          countKey += 1
          exercise.settingString[lMod*lLines+tempButton.id] = 1
        case mIndex to mIndex+7
          keys(countKey) = "mark"
          values(countKey) = str(tempbutton.id-mIndex+1)
          countKey += 1
          exercise.settingString[0] = bitset(exercise.settingString[0], tempButton.id-mIndex)
        case is >= 100
          lessions.add(tempButton.id)
          dim as integer i = 0
          foreach(long, lession, ks1)
            if lession = tempButton.id then exit while
            i += 1
          nexteach
          if i < ks1.size then exercise.settingString[i+1] = 1
      end select
    end if
  nexteach
  if countKey > 0 then redim preserve keys(countKey-1) else erase keys
  if lessions.size = 0 andalso countKey = 0 then
    exercise.currentList.list.addAll(exercise.fullList.list)
  else
    exercise.currentList.add , lessions, keys(), values(), Operators.or_
  end if
  exercise.updateLists
  if key = chr(255, 107) then return RETVAR.windowClosed
  return RETVAR.ok
end function

function conjugation(verb as TWord, inflection as integer) as TWord
  ' AT THE MOMENT, ONLY PRESENTE AND PRETÉRITO IS SUPPORTED
  dim as TWord ret
  if len(verb.word) < 2 then return ret    ' error
  dim as string retWord, retTrans, tempString, root, regularEnding
  dim as integer tempint = (inflection and 7)-1, mark = 1
  dim as string infl = ((tempint mod INFLECT.PLURAL) + 1) _
                       & iif(tempint\INFLECT.PLURAL, "p", "s"), reflex = ""
  if right(verb.word, 2) = "se" then
    select case as const inflection and 7
      case INFLECT.PERS1 + INFLECT.SINGULAR : reflex = "me "
      case INFLECT.PERS2 + INFLECT.SINGULAR : reflex = "te "
      case INFLECT.PERS3 + INFLECT.SINGULAR : reflex = "se "
      case INFLECT.PERS1 + INFLECT.PLURAL   : reflex = "nos "
      case INFLECT.PERS2 + INFLECT.PLURAL   : reflex = "os "
      case INFLECT.PERS3 + INFLECT.PLURAL   : reflex = "se "
    end select
  end if

  ' looking for irregular form
  select case as const inflection and (not 7)
    case INFLECT.PRES    : infl &= "pres"    : root = verb.map.get("rootpres")
    case INFLECT.PRETIND : infl &= "pretind" : root = verb.map.get("rootpretind")
    case INFLECT.PRETIMP : infl &= "pretimp" : root = verb.map.get("rootpretimp")
    case INFLECT.PERF    : infl &= "perf"    : root = verb.map.get("rootperf")
    case INFLECT.PQPERF  : infl &= "pqperf"  : root = verb.map.get("rootpqperf")
    case INFLECT.FUTSIMP : infl &= "futsimp" : root = verb.map.get("rootfutsimp")
    case INFLECT.FUTCOMP : infl &= "futcomp" : root = verb.map.get("rootfutcomp")
    case INFLECT.IMPERAT : infl &= "imperat"
  end select
  retWord  = verb.map.get("w" & infl)
  retTrans = verb.map.get("t" & infl)
  if retWord = "" then
    if len(root) = 0 then root = verb.map.get("root") : mark = 0
    ' regular form
    select case inflection and (not 7)
      ' PRESENTE DE INDICATIVO
      case INFLECT.PRES
        if (inflection and 7) <> INFLECT.PERS1+INFLECT.PLURAL and _
           (inflection and 7) <> INFLECT.PERS2+INFLECT.PLURAL then
          tempString = verb.map.get("rootdiph")
          if len(tempString) then root = tempString : mark = -1
        end if
        if len(root) = 0 then root = left(verb.word, len(verb.word)-1+2*(len(reflex)>0))
        tempInt = len(root)-1
        select case as const inflection and 7
          case INFLECT.PERS1 + INFLECT.SINGULAR
            regularEnding = "o"
          case INFLECT.PERS2 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "as" else regularEnding = "es"
          case INFLECT.PERS3 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "a" else regularEnding = "e"
          case INFLECT.PERS1 + INFLECT.PLURAL
            if root[tempInt] = 97 then
              regularEnding = "amos"
            elseif root[tempInt] = 101 then
              regularEnding = "emos"
            else
              regularEnding = "imos"
            end if
          case INFLECT.PERS2 + INFLECT.PLURAL
            if root[tempInt] = 97 then
              regularEnding = "áis"
            elseif root[tempInt] = 101 then
              regularEnding = "éis"
            else
              regularEnding = "ís"
            end if
          case INFLECT.PERS3 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "an" else regularEnding = "en"
        end select
      ' PRETÉRITO INDEFINIDO
      case INFLECT.PRETIND
        if len(root) = 0 then root = left(verb.word, len(verb.word)-1+2*(len(reflex)>0))
        tempInt = len(root)-1
        select case as const inflection and 7
          case INFLECT.PERS1 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "é"      else regularEnding = "í"
          case INFLECT.PERS2 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "aste"   else regularEnding = "iste"
          case INFLECT.PERS3 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "ó"      else regularEnding = "ió"
          case INFLECT.PERS1 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "amos"   else regularEnding = "imos"
          case INFLECT.PERS2 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "asteis" else regularEnding = "isteis"
          case INFLECT.PERS3 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "aron"   else regularEnding = "ieron"
        end select
      ' PRETÉRITO IMPERFECTO
      case INFLECT.PRETIMP
        if len(root) = 0 then root = left(verb.word, len(verb.word)-1+2*(len(reflex)>0))
        tempInt = len(root)-1
        select case as const inflection and 7
          case INFLECT.PERS1 + INFLECT.SINGULAR
            if len(root) = 0 then root = left(verb.word, len(verb.word)-1+2*(len(reflex)>0))
            tempInt = len(root)-1
            if root[tempInt] = 97 then regularEnding = "aba"    else regularEnding = "ía"
          case INFLECT.PERS2 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "abas"   else regularEnding = "ías"
          case INFLECT.PERS3 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "aba"    else regularEnding = "ía"
          case INFLECT.PERS1 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "ábamos" else regularEnding = "íamos"
          case INFLECT.PERS2 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "abais"  else regularEnding = "íais"
          case INFLECT.PERS3 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "aban"   else regularEnding = "ían"
        end select
    end select
    retWord = left(root, len(root)-1) & regularEnding
  end if
  retWord = reflex & retWord
  if retTrans = "" then retTrans = "<" & verb.translate.get(0) & ">"
  ret = TWord(retWord, retTrans, "verb")
  ret.mark = mark
  return ret
end function

sub getKeyArray(keys(any) as string)
  ' return an array with special keys (spanish and german)
  dim as integer S_COUNT = 21
  redim keys(S_COUNT-1)
  #ifdef FBGFX_FALLBACK
    ' TODO
    '============
    ' old values:
    '   { chr(160), chr(130), chr(161), chr(162), chr(163), _
    '     chr(164), chr(168), chr(142), chr(153), chr(154), chr(225), _
    '     chr(181), chr(144), chr(214), chr(244), chr(233), _
    '     chr(165)chr(173), chr(132), chr(148), chr(129) }
  #else
    keys( 0) = "á"
    keys( 1) = "é"
    keys( 2) = "í"
    keys( 3) = "ó"
    keys( 4) = "ú"
    keys( 5) = "ñ"
    keys( 6) = "¡"
    keys( 7) = "ä"
    keys( 8) = "ö"
    keys( 9) = "ü"
    keys(10) = "ß"
    keys(11) = "Á"
    keys(12) = "É"
    keys(13) = "Í"
    keys(14) = "Ó"
    keys(15) = "Ú"
    keys(16) = "Ñ"
    keys(17) = "¿"
    keys(18) = "Ä"
    keys(19) = "Ö"
    keys(20) = "Ü"
  #endif
end sub

sub guiAddKeys(byref tempGui as TGui, sID as integer, keys() as string, sx as integer, sy as integer, _
               w as integer, h as integer, dx as integer, dy as integer = 0, perLine as integer = 255)
  for i as integer = lbound(keys) to ubound(keys)
    tempGui.buttons.add TGuiButton(sID+i, sx+(w+dx)*(i mod perLine), sy+(h+dy)*(i\perLine), w, h, keys(i))
  next
end sub

sub guiAddLevs(byref tempGui as TGui, sID as integer, sx as integer, sy as integer, _
               w as integer, h as integer, dx as integer, fontsize as integer = 14)
  dim as string txt(...) = { "sehr schwer", "schwer", "eher schwer", "eher leicht", "leicht", "sehr leicht" }
  for i as integer = 0 to ubound(txt)
    dim as TGuiButton tempButton = TGuiButton(sID+i, sx+(w+dx)*i, sy, w, h, txt(i), chr(255, 62+i))
    tempButton.fontsize = fontsize
    tempButton.info = "F" & i+4
    tempGui.buttons.add tempButton
  next
end sub

sub init
  screenres SCREEN_X, SCREEN_Y, 32
  width SCREEN_X\8, SCREEN_Y\8
  color 0, &hffffff
  randomize
  windowtitle "Lengua v" & VERSION_STRING
  cls
  print "lade Daten ..."
end sub

function infoDefault(exercise as TExercise) as string
  dim as TList ptr listPtr = iif(exercise.currentList.size = 0, @exercise.fullList, @exercise.currentList)
  return "ausstehend: " & listPtr->notChecked & " von " & listPtr->size & " Wörtern"
end function

function infoEmpty(exercise as TExercise) as string
  return ""
end function

sub listToArray(l as TList, key as string, keyArray() as string, lpArray() as TListPoint)
  dim as TListPoint listPoint
  dim as TWord w
  dim as integer i = 0
  redim keyArray(l.list.size-1), lpArray(l.list.size-1)
  foreach(TListPoint, listPoint, l.list)
    w = globallist.get(listPoint.id)
    select case key
      case "word"
        keyArray(i) = modifyTxt(w.word, MODIFY_SORTABLE)
      case "mark"
        if w.mark then
          keyArray(i) = chr(w.mark)
        else
          keyArray(i) = chr(255)
        end if
      case "trans"
        keyArray(i) = modifyTxt(w.translate.get(0), MODIFY_SORTABLE)
      case "group"
        keyArray(i) = modifyTxt(w.groupname, MODIFY_SORTABLE)
      case "tag"
        keyArray(i) = modifyTxt(w.tag.get(0), MODIFY_SORTABLE)
      case "levelA"
        keyArray(i) = chr(w.levelA+12)
      case "levelB"
        keyArray(i) = chr(w.levelB+12)
      case "id"
        keyArray(i) = format(w.id, "###########")
      case else
        keyArray(i) = modifyTxt(w.map.get(key))
    end select
    lpArray(i) = listPoint
    i += 1
  nexteach
end sub

sub mainWindow
  ' choose and execute a test
  dim key as string, event as TGuiEvent, ret as integer
  ' buttons
  dim as integer mb_count = globalExercises.size
  dim as integer mb_startX = 50, mb_startY = 100, mb_width = 150, mb_height = 30, mb_dist = 20
  dim as integer mo_count = globalOptions.size
  dim as integer mo_startX = 550, mo_startY = 100, mo_width = 150, mo_height = 30, mo_dist = 20
  dim as TExercise exercise(mb_count-1), options(mo_count-1)
  dim as TList spec
  dim as long tempLong, filenumber = freefile
  dim as TGui mainGui
  dim as integer exIndex = 100, optIndex = 150, setIndex = 100

  open "spanish/exercise.dat" for binary as #filenumber
  for i as integer = 0 to mb_count-1
    exercise(i) = globalExercises.get(i)
    exercise(i).fullList = TList(exercise(i).listname)
    exercise(i).fullList.load
    if not eof(filenumber) then
      get #filenumber,, tempLong
      exercise(i).settingString = space(tempLong)
    end if
    if not eof(filenumber) then
      get #filenumber,, exercise(i).settingString
      exercise(i).settings(exercise(i), 0)
    end if
    mainGui.buttons.add( _
        TGuiButton(i+exIndex, mb_startX, mb_startY+(mb_height+mb_dist)*i, mb_width, mb_height, exercise(i).title))
    mainGui.buttons.add( _
        TGuiButton(i+exIndex+setIndex, exepath & "/img/settings.bmp", mb_startX+mb_width+mb_dist, _
        mb_startY+(mb_height+mb_dist)*i, mb_height, mb_height))
  next
  for i as integer = 0 to mo_count-1
    options(i) = globalOptions.get(i)
    options(i).fullList = TList(options(i).listname)
    options(i).fullList.load
    if not eof(filenumber) then
      get #filenumber,, tempLong
      options(i).settingString = space(tempLong)
    end if
    if not eof(filenumber) then
      get #filenumber,, options(i).settingString
      options(i).settings(options(i), 0)
    end if
    mainGui.buttons.add( _
        TGuiButton(i+optIndex, mo_startX, mo_startY+(mo_height+mo_dist)*i, mo_width, mo_height, options(i).title))
    mainGui.buttons.add( _
        TGuiButton(i+optIndex+setIndex, exepath & "/img/settings.bmp", mo_startX+mo_width+mo_dist, _
        mo_startY+(mo_height+mo_dist)*i, mo_height, mo_height))
  next
  close filenumber

  do
    ' draw GUI
    cls
    centerFT 400, 50, "Sprachtraining Spanisch - Deutsch", font, 18
    mainGui.show
    for i as integer = 0 to mb_count-1
      printFT mb_startX + mb_width + mb_height + 40, _
              mb_startY + i*(mb_height+mb_dist) + mb_height\2 - 6, _
              exercise(i).info(exercise(i)), font, 12
    next
    ' check input
    do
      key = inkey
      event = mainGui.getEvent
      if event.mButton = 1 then
        select case event.id
          case exIndex to exIndex+mb_count-1
            ret = exercise(event.id-exIndex).callback(exercise(event.id-exIndex))
            exercise(event.id-exIndex).fullList.lastSave = now
            exercise(event.id-exIndex).fullList.save
            exit do
          case exIndex+setIndex to exIndex+setIndex+mb_count-1
            ret = exercise(event.id-exIndex-setIndex).settings(exercise(event.id-exIndex-setIndex))
            exit do
          case optIndex to optIndex+mo_count-1
            ret = options(event.id-optIndex).callback(options(event.id-optIndex))
            exit do
        end select
      end if
      sleep 1
    loop until key = chr(255, 107) or ret = RETVAR.windowClosed
  loop until key = chr(255, 107) or ret = RETVAR.windowClosed
  open "spanish/exercise.dat" for binary as #filenumber
  for i as integer = 0 to mb_count-1
    put #filenumber,, cast(long, len(exercise(i).settingString))
    put #filenumber,, exercise(i).settingString
  next
  close filenumber
end sub

function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
  ' modify input - MODIFY_SORTABLE changes into lower case
  ' Using conflicting options may produce unexpected results
  dim as string ret
  dim as integer i = 0
  do until i >= len(txt)
    select case as const txt[i]
      case 0 to 31                                    ' ignore control chars
      case 65 to 90                                   ' upper case
        ret &= chr(txt[i] + iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), 32, 0))
      case 97 to 122                                  ' lower case
        ret &= chr(txt[i] - iif(opt and MODIFY_TOUPPER, 32, 0))
      case 40, 41                                     ' parentheses
        if opt and (MODIFY_NOPARANTHESES or MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
      case 32 to 39, 42 to 64, 91 to 96, 123 to 127   ' special keys
        if (opt and MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
      case 195
        i += 1
        if i >= len(txt) then exit do
        select case as const txt[i]
          case 132 ' Ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ae"
              elseif opt and MODIFY_TOUPPER then
                ret &= "AE"
              else
                ret &= "Ae"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ä", "Ä")
            end if
          case 164 ' ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "AE", "ae")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ä", "ä")
            end if
          case 150 ' Ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "oe"
              elseif opt and MODIFY_TOUPPER then
                ret &= "OE"
              else
                ret &= "Oe"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ö", "Ö")
            end if
          case 182 ' ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "OE", "oe")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ö", "ö")
            end if
          case 156 ' Ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ue"
              elseif opt and MODIFY_TOUPPER then
                ret &= "UE"
              else
                ret &= "Ue"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ü", "Ü")
            end if
          case 188 ' ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "UE", "ue")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ü", "ü")
            end if
          case 159 ' ß
            if opt and (MODIFY_SORTABLE or MODIFY_NOUMLAUT) then
              ret &= iif(opt and MODIFY_TOUPPER, "SS", "ss")
            else
              ret &= "ß"
            end if
          case 129 ' Á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "a", "A")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "á", "Á")
            end if
          case 161 ' á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "A", "a")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Á", "á")
            end if
          case 137 ' É
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "e", "E")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "é", "É")
            end if
          case 169 ' é
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "E", "e")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "É", "é")
            end if
          case 141 ' Í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "i", "I")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "í", "Í")
            end if
          case 173 ' í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "I", "i")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "´I", "´i")
            end if
          case 147 ' Ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "o", "O")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ó", "Ó")
            end if
          case 179 ' ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "O", "o")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ó", "ó")
            end if
          case 154 ' Ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "u", "U")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ú", "Ú")
            end if
          case 186 ' ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "U", "u")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ú", "ú")
            end if
          case 145 ' Ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOLOWER, "n|", "N|")
            elseif opt and MODIFY_NOTILDE then
              ret &= iif(opt and MODIFY_TOLOWER, "n", "N")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ñ", "Ñ")
            end if
          case 177 ' ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOUPPER, "N|", "n|")
            elseif opt and MODIFY_NOTILDE then
              ret &= iif(opt and MODIFY_TOUPPER, "N", "n")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ñ", "ñ")
            end if
          case else    ' in use: (192,161): ¡   (192,191): ¿
            if (opt and (MODIFY_SORTABLE or MODIFY_NOSPECIAL)) = 0 then ret &= "¡"
        end select
    end select
    i += 1
  loop
  return ret
end function

function number(first as integer = 0, last as integer = 0) as TWord
  ' create a random number from 0 to 999'999'999
  ' (greater numbers first will give back a wrong german translation)
  ' if "last" isn't greater than "first", give back the word for "first" (no random number)
  dim as integer symbol, no = iif(last > first, int(rnd*(last-first+1)) + first, first)
  dim as string retWord, retTrans
  dim as TWord ret, w, w1000, w100, w10
  static as TList list_numbers = TList("numbers")
  if list_numbers.list.isEmpty then list_numbers.add , 0, &hffffff, "group", "cardinal"

  ' numbers over 1 million
  if no > 999999 then
    w = searchWord(list_numbers, "1000000", "symbol")
    if no < 2000000 then
      retWord  = w.word
      retTrans = w.translate.get(0)
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no\1000000)
      retWord  = w.word & " " & retWord
      retTrans = w.translate.get(0) & " " & retTrans
    end if
    if no mod 1000000 > 0 then
      w = number(no mod 1000000)
      retWord  &= " " & w.word
      retTrans &= " " & w.translate.get(0)    
    end if
    ret = TWord(retWord, retTrans, "cardinal")
    ret.map.put("symbol", str(no))
    return ret
  end if

  ' numbers from 1'000 to 999'999
  if no > 999 then
    w = searchWord(list_numbers, "1000", "symbol")
    if no mod 1000 = 0 then
      retWord  = w.word
      retTrans = w.translate.get(0)
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no mod 1000)
      retWord  = replace(retWord,  "%", w.word)
      retTrans = replace(retTrans, "%", _
                 iif(no mod 1000 > 99 and no mod 1000 < 200, "ein", "") & w.translate.get(0))
    end if
    if no > 1999 then
      w = number(no\1000)
      retWord =  w.word & " " & retWord
      retTrans = w.translate.get(0) & retTrans
    end if
    ret = TWord(retWord, retTrans, "cardinal")
    ret.map.put("symbol", str(no))
    return ret
  end if

  ' numbers from 100 to 999
  if no > 99 then
    w = searchWord(list_numbers, no\100 & "00", "symbol")
    if no mod 100 = 0 then
      return w
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no mod 100)
      retWord  = replace(retWord,  "%", w.word)
      retTrans = replace(retTrans, "%", w.translate.get(0))
      ret = TWord(retWord, retTrans, "cardinal")
      ret.map.put("symbol", str(no))
      return ret
    end if
  end if

  ' numbers from 0 to 99
  w = searchWord(list_numbers, str(no), "symbol")
  if w.id then return w
  w = searchWord(list_numbers, no\10 & "0", "symbol")
  retWord  = w.map.get("generic-word")
  retTrans = w.map.get("generic-trans")
  w = number(no mod 10)
  retWord  = replace(retWord,  "%", w.word)
  retTrans = replace(retTrans, "%", w.translate.get(0))
  ret = TWord(retWord, retTrans, "cardinal")
  ret.map.put("symbol", str(no))
  return ret
end function

function searchInput(txt as string = "", x as integer = SCREEN_X\2 - 250, y as integer = SCREEN_Y\2-150, _
                     w as integer = 500, h as integer = 300) as string
  ' messagebox to enter words with support of special keys
  if x < 0 or y < 0 or w < 500 or h < 300 or x+w >= SCREEN_X or y+h >= SCREEN_Y then return txt
  dim as string keyString()
  getKeyArray keyString()
  dim searchGui as TGui, event as TGuiEvent
  dim as string key, ret = txt
  dim as integer mx, my, mb, maxLength = w\20
  dim as FB.Image ptr tempImg = imagecreate(w, h)
  get (x, y)-step(w-1, h-1), tempImg
  line (x, y)-step(w-1, h-1), &hffffff, bf
  line (x+1, y+1)-step(w-3, h-3), 0, b
  line (x+2, y+2)-step(w-5, h-5), 0, b
  line (x+4, y+4)-step(w-9, h-9), 0, b
  searchGui.buttons.add TGuiButton(1, x+w\2-150, y+130, 150, 30, "OK", chr(13))
  searchGui.buttons.add TGuiButton(2, x+w\2+30,  y+130, 150, 30, "Abbruch", chr(27))
  guiAddKeys searchGui, 3, keyString(), x+w\2-225, y+h-100, 30, 30, 12, 15, 11
  searchGui.show
  printFT x+30, y+40, "Eingabe: " & ret & "_", font, 15
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  do
    key = inkey
    event = searchGui.getEvent(key)
    if len(key) then
      select case key[0]
        case 8
          if len(ret) > 0 then
            dim as integer i = len(ret)-1
            do until ret[i] < 128 or ret[i] > 191 or i = 0
              i -= 1
            loop
            ret = left(ret, i)
            screenlock
            line (x+20, y+30)-step(w-30, 50), &hffffff, bf
            printFT x+30, y+40, "Eingabe: " & ret & "_", font, 15
            screenunlock
          end if
        case 31 to 128
          if len(ret) < maxLength then
            ret &= key
            screenlock
            line (x+20, y+30)-step(w-30, 50), &hffffff, bf
            printFT x+30, y+40, "Eingabe: " & ret & "_", font, 15
            screenunlock
          end if
      end select
    end if
    select case event.id
      case 1                         ' OK: return input
        exit do
      case 2                         ' Chancel: return old value
        ret = chr(27)
        exit do
      case 3 to ubound(keyString)+3  ' special key
        if len(ret) < maxLength then
          ret &= keyString(event.id-3)
          screenlock
          line (x+20, y+30)-step(w-30, 50), &hffffff, bf
          printFT x+30, y+40, "Eingabe: " & ret & "_", font, 15
          screenunlock
          do
            getmouse mx, my,, mb
            sleep 1
          loop until mb < 1
        end if
    end select
    sleep 1
  loop until key = chr(255, 107)
  put (x, y), tempImg, pset
  imagedestroy tempImg
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  return ret
end function

function simpleTest(exercise as TExercise) as long
  dim as string keyString()
  getKeyArray keyString()
  dim listpoint as TListPoint, w as TWord
  dim as integer solved, i = -1, done = 0, hint, col, count, mx, my, mb
  dim as string entry, key, firstTrans, tempString, article
  dim as FB.Image ptr okImg = imagecreate(16, 16)
  bload exepath & "/img/ok.bmp", okImg

  ' prepare GUI
  enum
    posNone = 0, posLevels
    posRepeat = 10, posHint, posShow, posSpeak, posHelp
    posKeys = 20
  end enum
  dim testGui as TGui, tempButton as TGuiButton
  dim as TGuiEvent event, eventNew
  ' special keys and levels of difficulty
  guiAddKeys testGui, posKeys, keyString(), 40, 320, 30, 30, 15, 15, 11
  guiAddLevs testGui, posLevels, 40, 450, 100, 40, 20
  ' button: repeat question
  tempButton = TGuiButton(posRepeat, 60, 530, 120, 50, "Wiederholen", chr(9), "Tab")
  tempButton.fontsize = 16
  testGui.buttons.add tempButton
  ' button: show hint
  tempButton = TGuiButton(posHint, 620, 530, 120, 50, "Hinweis")
  tempButton.fontsize = 16
  testGui.buttons.add tempButton
  ' button: show correct answer
  tempButton = TGuiButton(posShow, 330, 530, 120, 50, "ANZEIGEN", chr(13), "Enter")
  tempButton.fontsize = 16
  testGui.buttons.add tempButton
  ' button: speak word
  testGui.buttons.add _
      TGuiButton(posSpeak, exepath & "/img/speaker.bmp", 620, 320, 40, 40, chr(255, 60), "F2", GuiState.noBorder)
  ' button: show program help
  testGui.buttons.add _
      TGuiButton(posHelp, exepath & "/img/help.bmp", 690, 320, 40, 40, chr(255, 59), "F1", GuiState.noBorder)

  ' How many words shall be asked?
  for k as integer = 0 to abs(exercise.opt)-1
    if exercise.currentList.list.get(k).position > now then exit for
    count += 1
  next

  do until done >= count
    if done = count-1 then
      tempButton = testGui.getButton(posRepeat)
      tempButton.state or= GuiState.deactivated
      testGui.setButton(tempButton)
      tempButton.show
    end if
    i += 1
    solved = 0
    entry = ""
    listPoint = exercise.currentList.list.get(i-done)
    w = globallist.get(listPoint.id)
    if len(w.word) = 0 then
      exercise.fullList.list.remove(i-done)
      continue do
    end if
    hint = 0
    if exercise.opt > 0 then firstTrans = w.translate.get(0) else firstTrans = w.word
    for k as integer = 0 to ubound(keyString)
      tempButton = testGui.getButton(posKeys+k)
      tempButton.state and= not GuiState.deactivated
      testGui.setButton(tempButton)
    next
    tempButton = testGui.getButton(posHint)
    tempButton.state and= not GuiState.deactivated
    testGui.setButton(tempButton)
    screenlock
    cls
    testGui.show
    drawProgress 200, 610, 400, 10, done, count
    if exercise.opt > 0 then
      tempString = w.map.get("obj")      
      tempString = w.map.get("hint-es")
      printFT 100, 80, rtrim(w.word & " " & w.map.get("obj") _
                       & iif(len(tempString), " (" & tempString & ")", "")), font
    else
      tempString = w.map.get("hint-de")
      printFT 100, 80, w.translate.get(0) & iif(len(tempString), " (" & tempString & ")", ""), font
    end if
    printFT 100, 110, "Eingabe:", font
    printFT 190, 110, "_", font
    screenunlock
    do
      getmouse mx, my,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      if solved = 0 then
        select case asc(key)
          case 32 to 127
            if len(entry) < MAX_WORDLENGTH then
              entry &= key
            end if
          case 8
            dim as integer i = len(entry)-1
            do until i <= 0 orelse entry[i] < 128 or entry[i] > 191
              i -= 1
            loop
            entry = left(entry, i)
        end select
      end if
      getmouse mx, my,, mb
      if len(key) orelse mb > 0 then
        eventNew = testGui.getEvent(key)
      else
        eventNew.id = 0
      end if
      if eventNew.id <> event.id then
        event = eventNew
        select case event.id
          case posShow
            if solved = 0 then solved = 1
          case posSpeak
            if exercise.opt > 0 or solved = 2 then
              speakThread w.word
            end if
          case posRepeat
            exit do
          case posHint
            if solved = 0 then
              hint += 1
              do until hint >= len(firstTrans) orelse firstTrans[hint] < 128 or firstTrans[hint] > 191
                hint += 1
              loop
              if hint >= len(firstTrans) then solved = 1
            end if
          case posLevels to posLevels+5
            dim l as TListPoint, k as integer
            listPoint.modify(event.id-posLevels)
            listPoint.position = now + listPoint.addition - rnd/10
            ' update position in exercise.fullList and exercise.currentList
            for lno as integer = 0 to 1
              dim as TList ptr listPtr = iif(lno=0, @exercise.fullList, @exercise.currentList)
              dim as integer index = iif(lno=0, listPtr->list.indexOf(listPoint), i-done)
              for k = index+1 to listPtr->size - 1
                l = listPtr->list.get(k)
                if l.position > listPoint.position then exit for
              next
              if k = listPtr->size then
                listPtr->list.add(listPoint)
              else
                listPtr->list.add(k, listPoint)
              end if
              listPtr->list.remove(index)
              listPtr->update
            next
            done += 1
            exit do
          case posKeys to posKeys + ubound(keyString)
            if len(entry) < MAX_WORDLENGTH then entry &= keyString(event.id-posKeys)
          case posHelp
            dim as FB.Image ptr tempImg = imagecreate(SCREEN_X, SCREEN_Y)
            get (0, 0)-step(SCREEN_X-1, SCREEN_Y-1), tempImg
            help LANGUAGE & "/doc.htm", "simpleTest"
            put (0, 0), tempImg, pset
            imagedestroy tempImg
        end select
      end if
      if solved = 1 then
        for k as integer = 0 to ubound(keyString)
          tempButton = testGui.getButton(posKeys+k)
          tempButton.state or= GuiState.deactivated
          testGui.setButton(tempButton)
          tempButton.show
        next
        tempButton = testGui.getButton(posHint)
        tempButton.state or= GuiState.deactivated
        testGui.setButton(tempButton)
        tempButton.show
        ' special handling for nouns (colored output)
        if w.groupname = "noun" then
          tempString = iif(w.map.get("num") = "pl", "los ", "el ") 
          select case w.map.get("genus")
            case "f"
              article = w.map.get("article")
              if len(article) then
                tempString = article & " " & w.word
              else
                tempString = iif(tempstring = "el ", "la ", "las ") & w.word
              end if
              col = rgb(255, 0, 0)
            case "m"
              tempString &= w.word
              col = rgb(0, 0, 255)
            case else
              tempString = w.word
              col = 0
          end select
        else
          tempString = w.word
          col = 0
        end if
        screenlock
        line (100, 145)-step(600, 0)
        line (90, 146)-step(700, 35), rgb(255, 255, 255), bf
        if exercise.opt > 0 then
          printFT 100, 150, "Lösung:", font
          for k as integer = 0 to w.translate.size-1
            printFT 190, 150+25*k, w.translate.get(k), font
            if w.translate.get(k) = entry then put (80, 150), okImg, pset
          next
          if w.groupname = "noun" then
            line (100, 80)-step(600, 30), rgb(255, 255, 255), bf
            printFT 100, 80, tempString, font,, col
          end if
        else
          printFT 100, 150, "Lösung:", font
          printFT 190, 150, tempString, font,, col
          if tempString = entry or w.word = entry then put (80, 150), okImg, pset
        end if
        line (190, 105)-step(600, 35), rgb(255, 255, 255), bf
        printFT 190, 110, entry, font
        screenunlock
        solved = 2
      end if
      if solved = 0 andalso len(key) or event.id then
        screenlock
        line (190, 105)-step(610, 35), rgb(255, 255, 255), bf
        printFT 190, 110, entry & "_", font
        if hint > 0 then
          line (90, 146)-step(700, 35), rgb(255, 255, 255), bf
          printFT 100, 150, "Hinweis:", font,, &h777777
          printFT 190, 150, left(firstTrans, hint) & "...", font,, &h777777
        end if
        screenunlock
      end if
      if key = chr(255, 107) then exit do, do
      if key = chr(27) then
        select case message("Wollen Sie die Übung|wirklich abbrechen?", "Ja|Nein", 2)
          case 1
            exit do, do
          case RETVAR.windowClosed
            key = chr(255, 107)
            exit do, do
        end select
      end if
      sleep 1
    loop
    if i = count and done < i then i = done-1
  loop
  imagedestroy okImg
  select case key
    case chr(255, 107) : return RETVAR.windowClosed
    case chr(27)       : return RETVAR.aborted
    case else          : return RETVAR.ok
  end select
  return RETVAR.ok
end function

sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
  mutexlock speakLock
  dim search as integer, t as string = txt
  ready = -1
  mutexunlock speakLock
  search = instr(t, "ll")
  do while search
    t = left(t, search-1) & "y" & mid(t, search+2)
    search = instr(t, "ll")
  loop
  search = instr(t, "rr")
  do while search
    t = left(t, search) & "'" & mid(t, search+1)
    search = instr(t, "rr")
  loop
  shell "espeak -vmb-es" & no & " """ & t & """ -s100"
end sub

sub speakThread(txt as string, no as integer = 1)
  dim as integer ready = 0, r
  dim as any ptr thread = threadcall speak(txt, no, ready)
  do
    mutexlock speakLock
    r = ready
    mutexunlock speakLock
  loop until r
  threaddetach(thread)
end sub

sub wordInfo(w as TWord)
  dim as TWord tempWord
  if w.groupname = "verb" then
    dim as integer conjX = 50, conjPresX = 50, conjPresY = 100, conjPretX = 300, conjPretY = 100
    dim as integer conjImperfX = 550, conjImperfY = 100
    dim as integer conjHeadY = 50, conjLines = 20, conjSize = 14
    dim as string person(1 to 6) = { "yo", "tú", "él/ella", "nosotr.", "vosotr.", "ellos/-as" }
    screenlock
    cls
    dim as string tempString = w.translate.get(0)
    for i as integer = 1 to w.translate.size - 1
      tempString &= ", " & w.translate.get(i)
    next
    printFT conjX, conjHeadY, rtrim(w.word & " " & w.map.get("obj")) & "  -  " & tempString, font, conjSize
    printFT conjPresX, conjPresY, "presente de indicativo", font, conjSize
    for i as integer = 1 to 3
      tempWord = conjugation(w, INFLECT.PRES or i)
      printFT conjPresX, conjPresY+(i+.5)*conjLines, person(i), font, conjSize
      printFT conjPresX+80, conjPresY+(i+.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
      tempWord = conjugation(w, INFLECT.PRES or (i+INFLECT.PLURAL))
      printFT conjPresX, conjPresY+(i+3.5)*conjLines, person(i+3), font, conjSize
      printFT conjPresX+80, conjPresY+(i+3.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
    next
    printFT conjPretX, conjPretY, "pretérito de indicativo", font, conjSize
    for i as integer = 1 to 3
      tempWord = conjugation(w, INFLECT.PRETIND or i)
      printFT conjPretX, conjPretY+(i+.5)*conjLines, person(i), font, conjSize
      printFT conjPretX+80, conjPretY+(i+.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
      tempWord = conjugation(w, INFLECT.PRETIND or (i+INFLECT.PLURAL))
      printFT conjPretX, conjPretY+(i+3.5)*conjLines, person(i+3), font, conjSize
      printFT conjPretX+80, conjPretY+(i+3.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
    next
    printFT conjImperfX, conjImperfY, "pretérito de imperfecto", font, conjSize
    for i as integer = 1 to 3
      tempWord = conjugation(w, INFLECT.PRETIMP or i)
      printFT conjImperfX, conjImperfY+(i+.5)*conjLines, person(i), font, conjSize
      printFT conjImperfX+80, conjImperfY+(i+.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
      tempWord = conjugation(w, INFLECT.PRETIMP or (i+INFLECT.PLURAL))
      printFT conjImperfX, conjImperfY+(i+3.5)*conjLines, person(i+3), font, conjSize
      printFT conjImperfX+80, conjImperfY+(i+3.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
    next
    screenunlock
    getkey
  end if
end sub
